unit DeltaBlue;
(***************************************************************************
* DeltaBlue: This unit contains the source code for an implementation of the
* DeltaBlue incremental dataflow constraint solver written in Borland Object
* Pascal.
****************************************************************************
* Copyright (C) 2003 W&C Information Consultants
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* Contact the Author:
*
* W&C Information Consultants
* PO Box 11093
* Wierda Park South
* Centurion
* 0057
* South Africa
*
* tel.:   +27-12-653-5999
* fax.:   +27-12-653-5737
* mailto: willo@wack.co.za
*
***************************************************************************)

interface
uses
  Classes, SysUtils, variants;

const NO_METHOD = -1;

type
  TStrength = ( stRequired, stStrongPreferred, stPreferred, stStrongDefault,
                stDefault, stWeakDefault, stWeakest );
  TConstraint = class;
  TConstraints = class;
  TConstraintVariable = class;
  TConstraintVariables = class;
  TConstraintPlan = class;

  TConstraintEngine = class
  private
    FVariables: TConstraintVariables;
    FCurrentMark: longint;
  protected
    procedure NewMark;
  public
    constructor Create;
    destructor Destroy; override;

    function CreateVariable( Name : string; Value : Variant ) : TConstraintVariable;
    function CreateConstant( Name : string; Value : Variant ) : TConstraintVariable;

    function AddConstraint( C : TConstraint ) : TConstraint;
    procedure RemoveConstraint( C : TConstraint );

    procedure IncrementalRemove( C : TConstraint );
    procedure IncrementalAdd( C : TConstraint );

    property Variables : TConstraintVariables read FVariables;
    property CurrentMark : longint read FCurrentMark;
  end;

  (*
   \TConstraintVariable
   Constraint Variables are the glue of constraint graphs. In addition to its
   value, a variable has a set of all the constraints that refer to it and a
   pointer to the constraint that determines its value in the current constraint
   graph. If no constraint determines the variables value, the DeterminedBy
   property is set to nil. The property WalkStrength and Mark are used in method
   selection and plan extraction. The Stay property is used for constant
   propagation. This property is True id the variable's value was precomputed at
   planning time and need not be recomputed at plan execution time.
  *)
  TConstraintVariable = class
  private
    FValue: Variant;
    FConstraints: TConstraints;
    FStay: boolean;
    FMark: longint;
    FDeterminedBy: TConstraint;
    FWalkStrength: TStrength;
    FName: string;
    FEngine: TConstraintEngine;
    procedure SetConstraints(const Value: TConstraints);
    procedure SetDeterminedBy(const Value: TConstraint);
    procedure SetMark(const Value: integer);
    procedure SetStay(const Value: boolean);
    procedure SetWalkStrength(const Value: TStrength);
  public
    (* Variables are initialized at creation time as if they were determined by
       a virtual Stay constraint with a strength of stWeakest *)
    constructor Create( Engine : TConstraintEngine; Value : Variant );
    destructor Destroy; override;

    procedure SetValue(const Value: Variant);
    
    property Value : Variant read FValue write FValue;
    property Constraints : TConstraints read FConstraints write SetConstraints;
    property DeterminedBy : TConstraint read FDeterminedBy write SetDeterminedBy;
    property WalkStrength : TStrength read FWalkStrength write SetWalkStrength;
    property Mark : longint read FMark write SetMark;
    property Stay : boolean read FStay write SetStay;
    property Name : string read FName write FName;
    property Engine : TConstraintEngine read FEngine;
  end;

  TConstraintVariables = class(TList)
  private
    function GetItems(Index: Integer): TConstraintVariable;
    procedure SetItems(Index: Integer; const Value: TConstraintVariable);
  public
    property Items[Index:Integer] : TConstraintVariable read GetItems write SetItems; default;
  end;

  TConstraints = class(TList)
  private
    function GetItems(Index: Integer): TConstraint;
    procedure SetItems(Index: Integer; const Value: TConstraint);
  public
    function Pop : TConstraint;
    property Items[Index:Integer] : TConstraint read GetItems write SetItems; default;
  end;

  TConstraintPlan = class(TConstraints)
  private
    FEngine: TConstraintEngine;
  protected
    procedure Makeplan( Constraints : TConstraints );
  public
    constructor Create( Engine : TConstraintEngine ); overload;
    constructor Create( Engine : TConstraintEngine; Constraints : TConstraints ); overload;
    constructor Create( Engine : TConstraintEngine; Constraint : TConstraint ); overload;
    procedure Execute;

    property Engine : TConstraintEngine read FEngine;
  end;
  
  (*
    \TConstraint
    A TConstraint represents a constraint on a set of variables. The
    Methods property contains the set of alternative methods for enforcing it.
    The SelectedMethod property is used to record which of these methods is used
    to enforce the constraint in the current solution graph. If the constraint
    is not enforced, the SelectedMethod property is set to nil.

    The InputFlag property is True for constraints that depend on the outside
    world, such as a mouse constraint that sets a variable to the current mouse
    position.
  *)
  TOutputMap = array of integer;
  TConstraint = class
  private
    FInputFlag: Boolean;
    FSelectedMethod: integer;
    FVariables: TConstraintVariables;
    FStrength: TStrength;
    FEngine: TConstraintEngine;
    function getOutputMap(Index: integer): integer;
    function GetOutputs(Index: integer): TConstraintVariable;
    procedure SetInputFlag(const Value: Boolean);
    procedure setOutputMap(Index: integer; const Value: integer);
    procedure SetOutputs(Index: integer; const Value: TConstraintVariable);
    procedure SetSelectedMethod(const Value: integer);
    procedure SetStrength(const Value: TStrength);
    procedure SetVariables(const Value: TConstraintVariables);
    procedure setEngine(const Value: TConstraintEngine);
  protected
    FOutputMap : TOutputMap;
    function ChooseMethod : integer;
    function AddPropagate : boolean;
    procedure Recalculate;
  public
    constructor Create( aStrength : TStrength );
    destructor Destroy; override;

    procedure Execute; virtual;
    function MethodCount : integer; virtual;

    function InputsKnown : Boolean;
    function Satisfied : boolean;
    function Satisfy : TConstraint;

    property Variables : TConstraintVariables read FVariables write SetVariables;
    property Strength : TStrength read FStrength write SetStrength;
    property InputFlag : Boolean read FInputFlag write SetInputFlag;
    property SelectedMethod : integer read FSelectedMethod write SetSelectedMethod;
    property Outputs[Index: integer] : TConstraintVariable read GetOutputs write SetOutputs;
    property OutputMap[Index: integer] : integer read getOutputMap write setOutputMap;
    property Engine: TConstraintEngine read FEngine write setEngine;
  end;

  (*
    \TEditConstraint
    Edit variable v 
  *)
  TEditConstraint = class( TConstraint )
  public
    constructor Create( v : TConstraintVariable; aStrength : TStrength );
  end;

  (*
    \TStayConstraint
    Keep variable v constant
  *)
  TStayConstraint = class( TConstraint )
  public
    constructor Create( v : TConstraintVariable; aStrength : TStrength );
  end;

  (*
    \TEqualsConstraint
    v1 = v2
  *)
  TEqualsConstraint = class( TConstraint )
  public
    constructor Create( v1, v2 : TConstraintVariable; aStrength : TStrength );
    procedure Execute; override;
  end;

  (*
    \TSumConstraint
    v1 + v2 = sum
  *)
  TSumConstraint = class( TConstraint )
  public
    constructor Create( v1, v2, sum : TConstraintVariable; aStrength : TStrength );
    procedure Execute; override;
  end;

  (*
    \TMultiplyConstraint
    v1 * v2 = prod
  *)
  TMultiplyConstraint = class( TConstraint )
  public
    constructor Create( v1, v2, prod : TConstraintVariable; aStrength : TStrength );
    procedure Execute; override;
  end;

  (*
    \TScaleOffsetConstraint
    (src * scale) + offset = dest
  *)
  TScaleOffsetConstraint = class( TConstraint )
  public
    constructor Create( src, scale, offset, dest : TConstraintVariable; aStrength : TStrength );
    procedure Execute; override;
  end;

implementation

{ Utility Functions }
function NextDownstreamConstraint( Todo : TConstraints; Variable : TConstraintVariable ) : TConstraint;
var
  I : integer;
  C : TConstraint;
begin
  Result := nil;
  for I := 0 to Variable.Constraints.Count-1 do
  begin
    C := Variable.Constraints[I];
    if (C <> Variable.DeterminedBy) and
        C.Satisfied then
    begin
       if Result = nil then
         Result := C
       else
         Todo.Add( C );
    end;
  end;

  if Result = nil then
    Result := Todo.Pop;
end;

{ TConstraintVariable }

constructor TConstraintVariable.Create(Engine : TConstraintEngine; Value: Variant);
begin
  FValue := Value;
  FConstraints := TConstraints.Create;
  FDeterminedBy := nil;
  FWalkStrength := stWeakest;
  FMark := 0;
  FStay := True;
  FEngine := Engine;
end;

destructor TConstraintVariable.Destroy;
var
  I : integer;
begin
  for I := 0 to FConstraints.Count-1 do
    FConstraints[I].Variables.Remove( Self );
  FEngine.Variables.Remove( Self );
  FConstraints.Free;
  inherited;
end;

procedure TConstraintVariable.SetConstraints(const Value: TConstraints);
var
  I : integer;
begin
  FConstraints.Clear;
  for I := 0 to Value.Count-1 do
    FConstraints.Add( Value[I] );
end;

procedure TConstraintVariable.SetDeterminedBy(const Value: TConstraint);
begin
  FDeterminedBy := Value;
end;

procedure TConstraintVariable.SetMark(const Value: integer);
begin
  FMark := Value;
end;

procedure TConstraintVariable.SetStay(const Value: boolean);
begin
  FStay := Value;
end;

procedure TConstraintVariable.SetValue(const Value: Variant);
var
  C : TConstraint;
  Plan : TConstraintPlan;
begin
  C := Engine.AddConstraint( TEditConstraint.Create( self, stRequired ) );
  try
    if ( C.Satisfied ) then
    begin
      FValue := Value;
      Plan := TConstraintPlan.Create( Engine, C );
      try
        Plan.Execute;
      finally
        Plan.Free;
      end;
    end;
  finally
    Engine.RemoveConstraint( C );
    C.Free;
  end;
end;

procedure TConstraintVariable.SetWalkStrength(const Value: TStrength);
begin
  FWalkStrength := Value;
end;

{ TConstraints }

function TConstraints.GetItems(Index: Integer): TConstraint;
begin
  Result := TConstraint(inherited Items[Index]);
end;

function TConstraints.Pop: TConstraint;
begin
  Result := nil;
  if ( Count > 0 ) then
  begin
    Result := Items[0];
    Delete(0);
  end;
end;

procedure TConstraints.SetItems(Index: Integer; const Value: TConstraint);
begin
  inherited Items[Index] := Value;
end;

{ TConstraintVariables }

function TConstraintVariables.GetItems(
  Index: Integer): TConstraintVariable;
begin
  Result := TConstraintVariable(inherited Items[Index]);
end;

procedure TConstraintVariables.SetItems(Index: Integer;
  const Value: TConstraintVariable);
begin
  inherited Items[Index] := Value;
end;

{ TConstraint }

function TConstraint.AddPropagate: boolean;
var
  Todo : TConstraints;
  NextC: TConstraint;
  Output: TConstraintVariable;
begin
  Todo := TConstraints.Create;  (* unprocessed constraints *)
  try
    NextC := Self;
    while assigned(NextC) do
    begin
      Output := NextC.Outputs[NextC.SelectedMethod];
      if (Output.Mark = Engine.CurrentMark) then
      begin
        (* remove the cycle-causing constraint *)
        Engine.IncrementalRemove( Self );
        Result := False;
        exit;
      end;
      NextC.Recalculate;
      NextC := NextDownstreamConstraint(todo, output);
    end;
  finally
    Todo.Free;
  end;
  Result := true;
end;

function TConstraint.ChooseMethod;
var
  bestOutStrength : TStrength;
  m : integer;
  mOut : TConstraintVariable;
begin
  Result := NO_METHOD;
  bestOutStrength := Strength;
  for m := MethodCount - 1 downto 0 do
  begin
    mOut := Outputs[m];
    if ( mOut.Mark <> Engine.CurrentMark ) and
       ( mOut.WalkStrength > bestOutStrength ) then
    begin
      Result := m;
      bestOutStrength := mOut.WalkStrength;
    end;
  end;
end;

constructor TConstraint.Create( aStrength : TStrength );
begin
  FVariables := TConstraintVariables.Create;
  FInputFlag := false;
  FSelectedMethod := NO_METHOD;
  FStrength := aStrength;
  SetLength( FOutputMap, 0 );
end;

destructor TConstraint.Destroy;
var
  I : integer;
begin
  for I := 0 to FVariables.Count-1 do
    FVariables[I].Constraints.Remove( Self );
  FVariables.Free;
  SetLength( FOutputMap, 0 );
  inherited;
end;

procedure TConstraint.Execute;
begin
  (* default execute procedure; does nothing *)
end;

function TConstraint.getOutputMap(Index: integer): integer;
begin
  Result := FOutputMap[Index];
end;

function TConstraint.GetOutputs(Index: integer): TConstraintVariable;
begin
  Result := Variables[ FOutputMap[Index] ];
end;

function TConstraint.InputsKnown: Boolean;
var
  OutIndex, I : integer;
  Input : TConstraintVariable;
begin
  OutIndex := OutputMap[SelectedMethod];
  Result := true;
  for I := 0 to Variables.Count-1 do
  begin
    if ( I <> OutIndex ) then
    begin
      Input := Variables[I];
      if ( Input.Mark <> Engine.CurrentMark ) and
         not Input.Stay and
         ( Input.DeterminedBy <> nil ) then
      begin
        Result := false;
        break;
      end;
    end;
  end;
end;

function TConstraint.MethodCount: integer;
begin
  Result := length( FOutputMap );
end;

procedure TConstraint.Recalculate;

  function OutputWalkStrength : TStrength;
  var
    outIndex, m, mOutIndex : integer;
  begin
    Result := Strength;
    outIndex := OutputMap[ SelectedMethod ];
    for m := MethodCount-1 downto 0 do
    begin
      mOutIndex := OutputMap[m];
      if (mOutIndex <> outIndex) and
         (Variables[mOutIndex].WalkStrength > Result) then
        Result := Variables[mOutIndex].WalkStrength;
    end;
  end;

  function ConstantOutput : boolean;
  var
    outIndex, I : integer;
  begin
    Result := True;
    if ( InputFlag ) then
      Result := False
    else
    begin
      outIndex := OutputMap[SelectedMethod];
      for I := 0 to Variables.Count-1 do
      begin
        if I <> outIndex then
        begin
          if not Variables[I].Stay then
          begin
            Result := False;
            break;
          end;
        end;
      end;
    end;
  end;

var
  Output : TConstraintVariable;
  
begin
  Output := Outputs[SelectedMethod];
  Output.WalkStrength := OutputWalkStrength;
  Output.Stay := ConstantOutput;
  if (Output.Stay) then
    Execute;
end;

function TConstraint.Satisfied: boolean;
begin
  Result := SelectedMethod <> NO_METHOD;
end;

function TConstraint.Satisfy: TConstraint;
var
  outIndex, I : integer;
  overridden : TConstraint;
  output : TConstraintVariable;
begin
  Result := nil;
  SelectedMethod := ChooseMethod();
  if Satisfied then
  begin
    (* mark inputs to allow cycle detection in AddPropagate *)
    outIndex := OutputMap[ SelectedMethod ];
    for i := 0 to Variables.Count-1 do
      if i <> outIndex then
        Variables[i].Mark := Engine.CurrentMark;

    output := Variables[outIndex];
    overridden := output.DeterminedBy;
    if (overridden <> nil) then
      overridden.SelectedMethod := NO_METHOD;
    output.DeterminedBy := Self;
    if (not AddPropagate) then
      raise Exception.Create( 'Cycle Encountered' );

    output.Mark := Engine.CurrentMark;
    Result := overridden;
  end
  else
  begin
    if Strength = stRequired then
      raise Exception.Create( 'Could not satisfy a required constraint ('+ClassName+')' );
  end;
end;

procedure TConstraint.setEngine(const Value: TConstraintEngine);
begin
  FEngine := Value;
end;

procedure TConstraint.SetInputFlag(const Value: Boolean);
begin
  FInputFlag := Value;
end;

procedure TConstraint.setOutputMap(Index: integer; const Value: integer);
begin
  FOutputMap[Index] := Value;
end;

procedure TConstraint.SetOutputs(Index: integer;
  const Value: TConstraintVariable);
var
  I : integer;
begin
  I := Variables.IndexOf( Value );
  FOutputMap[Index] := I;
end;

procedure TConstraint.SetSelectedMethod(const Value: integer);
begin
  FSelectedMethod := Value;
end;

procedure TConstraint.SetStrength(const Value: TStrength);
begin
  FStrength := Value;
end;

procedure TConstraint.SetVariables(const Value: TConstraintVariables);
var
  I : integer;
begin
  FVariables.Clear;
  for I := 0 to Value.Count-1 do
    FVariables.Add( Value[I] );
end;

{ TConstraintEngine }

function TConstraintEngine.AddConstraint(C: TConstraint) : TConstraint;
var
  Overridden : TConstraint;
begin
  C.Engine := Self;
  Result := C;
  NewMark();
  Overridden := C.Satisfy();
  while (overridden <> nil) do
    Overridden := Overridden.Satisfy();
end;

constructor TConstraintEngine.Create;
begin
  FVariables := TConstraintVariables.Create;
  FCurrentMark := 0;
end;

function TConstraintEngine.CreateConstant(Name: string;
  Value: Variant): TConstraintVariable;
begin
  Result := TConstraintVariable.Create( Self, Value );
  Result.Name := Name;
  Result.WalkStrength := stRequired;
  Result.Stay := True;
  Variables.Add( Result );
end;

function TConstraintEngine.CreateVariable(Name: string;
  Value: Variant): TConstraintVariable;
begin
  Result := TConstraintVariable.Create( Self, Value );
  Result.Name := Name;
  Result.Stay := True;
  Variables.Add( Result );
end;

destructor TConstraintEngine.Destroy;
begin
  while FVariables.Count > 0 do
    FVariables[0].Free;

  FVariables.Free;
  inherited;
end;

procedure TConstraintEngine.IncrementalAdd(C: TConstraint);
begin
  AddConstraint( C );
end;

procedure TConstraintEngine.IncrementalRemove(C: TConstraint);
var
  Output : TConstraintVariable;
  Unsatisfied : TConstraints;
  I : integer;

  procedure RemovePropagateFrom( v : TConstraintVariable );
  var
    nextC : TConstraint;
    todo  : TConstraints;
    i     : integer;
  begin
    todo := TConstraints.Create;
    try
      v.DeterminedBy := nil;
      v.WalkStrength := stWeakest;
      v.Stay := true;

      while true do
      begin
        for i := 0 to v.Constraints.Count-1 do
          if (not v.Constraints[i].Satisfied) then
            Unsatisfied.Add(c);
        nextC := NextDownstreamConstraint(todo, v);
        if nextC = nil then
          break
        else
        begin
          nextC.Recalculate;
          v := nextC.Outputs[nextC.SelectedMethod];
        end;
      end;
    finally
      todo.Free;
    end;
  end;
  
var
  aStrength : TStrength;
begin
  Output := C.Outputs[C.SelectedMethod];
  C.SelectedMethod := NO_METHOD;
  for I := 0 to C.Variables.Count-1 do
    C.Variables[I].Constraints.Remove( C );

    Unsatisfied := TConstraints.Create;
    try
      RemovePropagateFrom( Output );

      for aStrength := high(TStrength) downto low(TStrength) do
      begin
        for I := 0 to Unsatisfied.Count-1 do
        begin
          if Unsatisfied[I].Strength = aStrength then
          begin
            IncrementalAdd( Unsatisfied[I] );
          end;
        end;
      end;
    finally
      Unsatisfied.Free;
    end;
end;

procedure TConstraintEngine.NewMark;
begin
  inc( FCurrentMark );
end;

procedure TConstraintEngine.RemoveConstraint(C: TConstraint);
begin
  if C.Satisfied then
  begin
    IncrementalRemove(C);
  end;
end;

{ TConstraintPlan }

constructor TConstraintPlan.Create(Engine : TConstraintEngine );
var
  I, J : integer;
  Constraints: TConstraints;
  Constraint : TConstraint;
begin
  FEngine := Engine;
  Constraints := TConstraints.Create;
  try
    for I := 0 to Engine.Variables.Count-1 do
    begin
      for J := 0 to Engine.Variables[I].Constraints.Count-1 do
      begin
        Constraint := Engine.Variables[I].Constraints[J];
        if Constraint.InputFlag and
           Constraint.Satisfied then
        begin
          Constraints.Add( Constraint );
        end;
      end;
    end;
    Makeplan( Constraints );
  finally
    Constraints.Free;
  end;
end;

constructor TConstraintPlan.Create(Engine: TConstraintEngine; Constraints: TConstraints);
var
  I : integer;
begin
  FEngine := Engine;
  I := 0;
  while I < Constraints.Count do
  begin
    if Constraints[I].InputFlag and Constraints[I].Satisfied then
      inc(I)
    else
      Constraints.Delete(I);
  end;
  Makeplan( Constraints );
end;

constructor TConstraintPlan.Create(Engine: TConstraintEngine; Constraint: TConstraint);
var
  Constraints: TConstraints;
begin
  FEngine := Engine;
  Constraints := TConstraints.Create;
  try
    if Constraint.InputFlag and
       Constraint.Satisfied then
    begin
      Constraints.Add( Constraint );
      Makeplan( Constraints );
    end;
  finally
    Constraints.Free;
  end;
end;

procedure TConstraintPlan.Execute;
var
  I : integer;
begin
  for I := 0 to Count-1 do
  begin
    Items[I].Execute;
  end;
end;

procedure TConstraintPlan.Makeplan(Constraints: TConstraints);
var
  NextC : TConstraint;
  Output: TConstraintVariable;
begin
  Engine.NewMark();
  NextC := Constraints.Pop;
  while assigned( NextC ) do
  begin
    Output := NextC.Outputs[NextC.SelectedMethod];
    if (Output.Mark <> Engine.CurrentMark) and
       NextC.InputsKnown then
    begin
      Add( NextC );
      Output.Mark := Engine.CurrentMark;
      NextC := NextDownstreamConstraint( Constraints, Output );
    end
    else
      NextC := Constraints.Pop;
  end;
end;

{ TEditConstraint }

constructor TEditConstraint.Create(v: TConstraintVariable;
  aStrength: TStrength);
begin
  inherited Create( aStrength );
  InputFlag := True;
  Variables.Add( v );
  v.Constraints.Add( Self );
  SetLength( FOutputMap, 1 );
  OutputMap[0] := 0;
end;

{ TStayConstraint }

constructor TStayConstraint.Create( v : TConstraintVariable; aStrength : TStrength );
begin
  inherited Create( aStrength );
  Variables.Add( v );
  v.Constraints.Add( Self );
  SetLength( FOutputMap, 1 );
  OutputMap[0] := 0;
end;

{ TEqualsConstraint }

constructor TEqualsConstraint.Create(v1, v2: TConstraintVariable;
  aStrength: TStrength);
begin
  inherited Create( aStrength );
  Variables.Add( v1 );
  Variables.Add( v2 );
  v1.Constraints.Add( Self );
  v2.Constraints.Add( Self );
  SetLength( FOutputMap, 2 );
  OutputMap[0] := 0;
  OutputMap[1] := 1;
end;

procedure TEqualsConstraint.Execute;
begin
  case SelectedMethod of
    0 : Variables[0].Value := Variables[1].Value;
    1 : Variables[1].Value := Variables[0].Value;
  end;
end;

{ TSumConstraint }

constructor TSumConstraint.Create(v1, v2, sum: TConstraintVariable;
  aStrength: TStrength);
begin
  inherited Create( aStrength );
  Variables.Add( v1 );
  Variables.Add( v2 );
  Variables.Add( sum );
  v1.Constraints.Add( Self );
  v2.Constraints.Add( Self );
  sum.Constraints.Add( Self );
  SetLength( FOutputMap, 3 );
  OutputMap[0] := 2;
  OutputMap[1] := 1;
  OutputMap[2] := 0;
end;

procedure TSumConstraint.Execute;
begin
  (* a + b = sum *)
  case SelectedMethod of
    0 : Variables[2].Value := Variables[0].Value + Variables[1].Value;
    1 : Variables[1].Value := Variables[2].Value - Variables[0].Value;
    2 : Variables[0].Value := Variables[2].Value - Variables[1].Value;
  end;
end;

{ TMultiplyConstraint }

constructor TMultiplyConstraint.Create(v1, v2, prod: TConstraintVariable;
  aStrength: TStrength);
begin
  inherited Create( aStrength );
  Variables.Add( v1 );
  Variables.Add( v2 );
  Variables.Add( prod );
  v1.Constraints.Add( Self );
  v2.Constraints.Add( Self );
  prod.Constraints.Add( Self );
  SetLength( FOutputMap, 3 );
  OutputMap[0] := 2;
  OutputMap[1] := 1;
  OutputMap[2] := 0;
end;

procedure TMultiplyConstraint.Execute;
begin
  (* a * b = prod *)
  case SelectedMethod of
    0 : Variables[2].Value := Variables[0].Value * Variables[1].Value;
    1 : Variables[1].Value := Variables[2].Value / Variables[0].Value;
    2 : Variables[0].Value := Variables[2].Value / Variables[1].Value;
  end;
end;

{ TScaleOffsetConstraint }

constructor TScaleOffsetConstraint.Create(src, scale, offset,
  dest: TConstraintVariable; aStrength: TStrength);
begin
  inherited Create( aStrength );
  Variables.Add( src );
  Variables.Add( scale );
  Variables.Add( offset );
  Variables.Add( dest );
  src.Constraints.Add( Self );
  scale.Constraints.Add( Self );
  offset.Constraints.Add( Self );
  dest.Constraints.Add( Self );
  SetLength( FOutputMap, 4 );
  OutputMap[0] := 3;
  OutputMap[1] := 0;
end;

procedure TScaleOffsetConstraint.Execute;
begin
  (* (src * scale) + offset = dest *)
  case SelectedMethod of
    0 : Variables[3].Value := (Variables[0].Value * Variables[1].Value) + Variables[2].Value;
    1 : Variables[0].Value := (Variables[3].Value - Variables[2].Value) / Variables[1].Value;
  end;
end;

end.
